{-# LANGUAGE TupleSections #-}

module Eval where

import Util

import Abstract
import Value
import TCM

-- import CallStack
import TraceError

import Data.Maybe -- fromMaybe
import Data.List -- find
import Data.Map (Map)
import qualified Data.Map as Map

import Control.Monad.Identity
import Control.Monad.State
import Control.Monad.Error
import Control.Monad.Reader

import Debug.Trace


traceEta msg a = a -- trace msg a 
traceEtaM msg = return () -- traceM msg
{-
traceEta msg a = trace msg a 
traceEtaM msg = traceM msg
-}

traceMatch msg a = a -- trace msg a 
traceMatchM msg = return () -- traceM msg
{- 
traceMatch msg a = trace msg a 
traceMatchM msg = traceM msg
-}
 

-- evaluation with rewriting -------------------------------------

{- 

Rewriting rules have the form

  blocked --> pattern

this means that at the root, at most one rewriting step is possible.
Rewriting rules are considered computational, since they trigger new
(symbolic) computations.  At least they have to be applied in

- pattern matching
- equality checking

When a new rule b --> p is added, b should be in --> normal form.
Otherwise there could be inconsistencies, like adding both rules

  b --> true
  b --> false

If after adding b --> true b is rewritten to nf, then the second rule
would be true --> false, which can be captured by MiniAgda.

Also, after adding a new rule, it could be used to rewrite the old rules.

Implementation:

- add a set of local rewriting rules to the context (not to the state)
- keep values in --> weak head normal form
- untyped equality test between values

 -}

reval :: Val -> TypeCheck Val
reval u = -- trace ("reval " ++ show u) $ 
 case u of
  VSort (CoSet v) -> reval v >>= return . VSort . CoSet
  VSort{} -> return u
  VInfty  -> return u
  VSucc{} -> return u  -- no rewriting in size expressions
  VMax{}  -> return u
{-
  VSucc v -> do v' <- reval v
                return $ succSize v'
 -}
  VApp v vl          -> do
    v'  <- reval v
    vl' <- mapM reval vl
    w   <- foldM app v' vl'
    reduce w  -- since we only have rewrite rules at base types
              -- we do not need to reduces prefixes of w
--  VCon{} -> return u
  VDef{} -> return $ VApp u [] -- restore invariant
                               -- CAN'T rewrite defined fun/data
  VGen{} -> reduce u  -- CAN rewrite variable

  VCase v env cl     -> do v' <- reval v
                           env' <- reEnv env
                           evalCase v' env' cl

  VPi dec x av env b -> do av' <- reval av
                           env' <- reEnv env
                           return $ VPi dec x av' env' b

  VLam x env e       -> do env' <- reEnv env
                           return $ VLam x env' e

  VClos env e        -> do env' <- reEnv env
                           return $ VClos env' e

  VMeta{} -> return u
  VUp v tv           -> do v' <- reval v
                           tv' <- reval tv
                           up False v' tv'  -- do not force at this point
  VSing v tv         -> do v' <- reval v
                           tv' <- reval tv
                           return $ vSing v' tv'
  VIrr -> return u

-- This is a bit of a hack (finding a fresh name)
-- <t : Pi x:a.b> = Pi x:a <t x : b>
-- <t : <t' : a>> = <t' : a>
vSing :: Val -> TVal -> TVal
vSing v (VPi dec x av env b) = 
 VPi dec x' av (update env xv v) $ Sing (App (Var xv) (Var x')) b
  where xv = ("vSing#" ++ x)
        x' = if x == "" then fresh env else x   
vSing _ tv@(VSing{}) = tv  
vSing v tv = VSing v tv

reEnv :: Env -> TypeCheck Env
reEnv (Environ rho mmeas) = do
  rho' <- mapM (\ (x,v) -> reval v >>= return . (x,)) rho 
  return $ Environ rho' mmeas -- no need to reevaluate mmeas, since only sizes
   

-- reduce the root of a value
reduce :: Val -> TypeCheck Val
reduce v = -- trace ("reduce " ++ show v) $
 do
  rewrules <- asks rewrites
  mr <- findM (\ rr -> equal v (lhs rr)) rewrules
  case mr of
     Nothing -> return v
     Just rr -> traceRew ("firing " ++ show rr) $ return (rhs rr)

-- equal v v'  tests values for untyped equality
-- precond: v v' are in --> whnf
equal :: Val -> Val -> TypeCheck Bool
equal u1 u2 = -- trace ("equal " ++ show u1 ++ " =?= " ++ show u2) $
  case (u1,u2) of
    (v1,v2) | v1 == v2 -> return True -- includes all size expressions
--    (VSucc v1, VSucc v2) -> equal v1 v2  -- NO REDUCING NECC. HERE (Size expr)
    (VApp v1 vl1, VApp v2 vl2) -> 
       liftM2 (&&) (equal' v1 v2) (equals' vl1 vl2)
    (VPi dec1 x1 av1 env1 b1,VPi dec2 x2 av2 env2 b2) ->
       liftM2 (&&) (equal av1 av2) $     -- NO RED. NECC. (Type)
         new x1 (Domain av1 dec1) $ \ vx -> do 
                  v1 <- whnf (update env1 x1 vx) b1
                  v2 <- whnf (update env2 x2 vx) b2 
                  equal v1 v2

    (VSing v1 tv1, VSing v2 tv2) -> liftM2 (&&) (equal' v1 v2) (equal' tv1 tv2) 

    (VLam x1 env1 b1, VLam x2 env2 b2) -> -- PROBLEM: DOMAIN MISSING
         addName x1 $ \ vx -> do          -- CAN'T "up" fresh variable
               do v1 <- whnf (update env1 x1 vx) b1
                  v2 <- whnf (update env2 x2 vx) b2
                  equal' v1 v2
    _ -> return False
    
equals' :: [Val] -> [Val] -> TypeCheck Bool
equals' [] []             = return True
equals' (w1:vs1) (w2:vs2) = liftM2 (&&) (whnfClos w1 >>= \ v1 -> 
                                         whnfClos w2 >>= \ v2 -> equal' v1 v2) 
                                        (equals' vs1 vs2)
equals' vl1 vl2           = return False

equal' = equal

{- LEADS TO NON-TERMINATION
-- equal' v1 v2  tests values for untyped equality
-- v1 v2 are not necessarily in --> whnf
equal' v1 v2 = do
  v1' <- reduce v1
  v2' <- reduce v2
  equal v1' v2'
-}

-- normalization -----------------------------------------------------

reify :: Val -> TypeCheck Expr
reify v = reify' (5, True) v

-- normalize to depth m
reify' :: (Int, Bool) -> Val -> TypeCheck Expr
reify' m v0 = 
  case v0 of
    (VClos rho e)        -> whnf rho e >>= reify' m
    (VInfty)             -> return $ Infty
    (VSucc v)            -> liftM Succ $ reify' m v
    (VMax vs)            -> liftM Max  $ mapM (reify' m) vs
    (VMeta x n)          -> return $ iterate Succ (Meta x) !! n
    (VSort (CoSet v))    -> reify' m v >>= return . Sort . CoSet
    (VSort s)            -> return $ Sort $ vSortToSort s
    (VPi dec x va rho e) -> do
          a <- reify' m va
          newWithGen x (Domain va dec) $ \ k xv -> do
            vb <- whnf (update rho x xv) e
            liftM (Pi (TBind dec (x++"~"++show k) a)) $ reify' m vb
    (VSing v tv)         -> liftM2 Sing (reify' m v) (reify' m tv)
    (VLam x rho e)       -> do
          addName x $ \ xv@(VGen k) -> do
            vb <- whnf (update rho x xv) e
            liftM (Lam defaultDec (x++"~"++show k)) $ reify' m vb  -- TODO: dec!?
    (VUp v tv)           -> reify' m v -- TODO: type directed reification
    (VGen k)             -> return $ Var $ "~" ++ show k
    (VDef d)             -> return $ Def d
    (VApp v vl)          -> if fst m > 0 && snd m 
                             then force v0 >>= reify' (fst m - 1, True) -- forgotten the meaning of the boolean, WAS: False) 
                             else let m' = (fst m, True) in
                               liftM2 (foldl App) (reify' m' v) (mapM (reify' m') vl)
    (VCase v rho cls)    -> do
          e <- reify' m v
          return $ Case e cls -- TODO: properly evaluate clauses!!
    (VIrr)               -> return $ Irr
  

-- evaluation --------------------------------------------------------

-- weak head normal form
-- monadic, since it reads the globally defined constants from the signature
-- lets are expanded away

whnf :: Env -> Expr -> TypeCheck Val
whnf env e = enter ("whnf " ++ show e) $
  case e of
    Meta i -> return $ VMeta i 0
    LLet (TBind dec x t1) e1 e2 -> do 
      let v1 = mkClos env e1
      whnf (update env x v1) e2
{-
-- ALT: remove erased lambdas entirely
    Lam dec x e1 | erased dec -> whnf env e1
                 | otherwise -> return $ VLam x env e1
-}
    Lam dec x e1 -> return $ VLam x env e1
    Pi (TBind dec x a) b -> do 
      av <- whnf env a  -- Pi is strict in its first argument
      return $ VPi dec x av env b

    -- a measured type evaluates to 
    -- * a bounded type if measure present in environment (rhs of funs)
    -- * otherwise to a measured type (lhs of funs)
    Pi (TMeasure mu) b -> do
      muv <- whnfMeasure env mu
      bv  <- whnf env b -- not adding measure constraint to context!
      case (envBound env) of
        Nothing   -> return $ VMeasured muv bv 
           -- fail $ "panic: whnf " ++ show e ++ " : no measure in environment " ++ show env
        Just muv' -> return $ VGuard (Bound muv muv') bv

    Pi (TBound (Bound mu mu')) b -> do
          muv  <- whnfMeasure env mu  
          muv' <- whnfMeasure env mu'
          bv   <- whnf env b  -- not adding measure constraint to context!
          return $ VGuard (Bound muv muv') bv

    Sing e t  -> do tv <- whnf env t
                    sing env e tv
{-
-- ALT: filter out all erased arguments from application
    App e1 el -> do v1 <- whnf env e1
                    vl <- liftM (filter (/= VIrr)) $ mapM (whnf env) el
                    app v1 vl
-}
    App f e   -> do vf <- whnf env f
                    let ve = mkClos env e
                    app vf ve
{-
    App e1 el -> do v1 <- whnf env e1
                    vl <- mapM (whnf env) el
                    app v1 vl
-}

    Case e cs -> do v  <- whnf env e
                    evalCase v env cs
                  -- trace ("case head evaluates to " ++ showVal v) $ return ()

    Sort s -> whnfSort env s >>= return . VSort
    Infty -> return VInfty
    Zero -> return VZero
    Succ e1 -> do v <- whnf env e1           -- succ is strict
                  return $ succSize v

    Max es  -> do vs <- mapM (whnf env) es   -- max is strict
                  return $ maxSize vs
               
    Def (DefId Let n) -> do 
        (LetSig _ v) <- lookupSymb n
        whnfClos v

    Def id   -> return $ vDef id
{-
    Con co n -> return $ VCon co n
    
    Def n -> return $ VDef n

    Let n -> do sig <- gets signature 
                let (LetSig _ v) = lookupSig n sig
                return v
--                let (LetSig _ e) = lookupSig n sig
--                whnf [] e
-}
    Var y -> lookupEnv env y >>= whnfClos
    Erased e -> whnf env e -- return VIrr -- NEED TO KEEP because of eta-exp!
    Irr -> return VIrr

whnfMeasure :: Env -> Measure Expr -> TypeCheck (Measure Val)
whnfMeasure rho (Measure mu) = mapM (whnf rho) mu >>= return . Measure

whnfSort :: Env -> Sort Expr -> TypeCheck (Sort Val)
whnfSort rho (CoSet e) = whnf rho e >>= return . CoSet
whnfSort rho (Set e) = whnf rho e >>= return . Set
whnfSort rho Type = return Type
whnfSort rho Size = return Size
whnfSort rho Tm = return Tm

whnfClos :: Clos -> TypeCheck Val
whnfClos v = -- trace ("whnfClos " ++ show v) $
  case v of
    (VClos e rho) -> whnf e rho
    (VApp (VDef (DefId Fun n)) vl) -> appDef n vl -- THIS IS TO SOLVE A PROBLEM
    v -> return v
{- THE PROBLEM IS that
  (tail (x Up Stream)) Up Stream is a whnf, because Up Stream is lazy
  in equality checking this is a problem when the Up is removed.
-}

-- evaluate in standard environment
whnf' :: Expr -> TypeCheck Val
whnf' e = do
  env <- getEnv
  whnf env e

-- <t : Pi x:a.b> = Pi x:a <t x : b>
-- <t : <t' : a>> = <t' : a>
sing :: Env -> Expr -> TVal -> TypeCheck TVal
sing rho e tv = do
  let v = mkClos rho e -- v <- whnf rho e
  return $ vSing v tv
{-
sing env' e (VPi dec x av env b)  = do 
  return $ VPi dec x' av env'' (Sing (App e (Var x')) b)
    where env'' = env' ++ env  -- super ugly HACK
          x'    = if x == "" then fresh env'' else x   
    -- Should work with just x since shadowing is forbidden
sing _ _ tv@(VSing{}) = return $ tv
sing env e tv = do v <- whnf env e      -- singleton strict, is this OK?!
                   return $ VSing v tv 
-}

sing' :: Expr -> TVal -> TypeCheck TVal
sing' e tv = do
  env <- getEnv
  sing env e tv

evalCase :: Val -> Env -> [Clause] -> TypeCheck Val
evalCase v env cs = do
  m  <- matchClauses env cs [v]
  case m of
    Nothing -> return $ VCase v env cs
    Just v' -> return $ v'

piApp :: TVal -> Clos -> TypeCheck TVal
piApp (VPi dec x av env b) w = whnf (update env x w) b
-- piApp _ = fail "piApp: IMPOSSIBLE"

piApps :: TVal -> [Clos] -> TypeCheck TVal
piApps tv [] = return tv
piApps tv (v:vs) = do tv' <- piApp tv v
                      piApps tv' vs

-- in app u v, u might be a VDef (e.g. when coming from reval)
app :: Val -> Clos -> TypeCheck Val
app u v = do  
         case u of
            VDef (DefId Fun n) -> appDef n [v]
            VApp (VDef (DefId Fun n)) vl -> appDef n (vl ++ [v])
            VApp h@(VDef (DefId (Con Ind) n)) vl -> do
              v <- whnfClos v      -- inductive constructors are strict!
              return $ VApp h (vl ++ [v])
--            VDef n -> appDef n [v]
--            VApp (VDef id) vl -> VApp (VDef id) (vl ++ [v])
            VApp v1 vl -> return $ VApp v1 (vl ++ [v])
            VLam x env e  -> whnf (update env x v) e
            VUp u1 (VPi dec x av rho b) -> do
{- 
-- ALT: erased functions are not applied to their argument!
              v1 <- if erased dec then return v else app v [w]  -- eta-expand w ??
-}
              v1 <- app u1 v  -- eta-expand v ??
              bv <- whnf (update rho x v) b
              up False v1 bv
              
{-
-- ALT: VIrr consumes applications
            (VIrr,_) -> return VIrr
 -}
            VIrr -> fail $ "app internal error: " ++ show (VApp u [v])
            _ -> return $ VApp u [v]
-- 
-- app :: Val -> [Val] -> TypeCheck Val
-- app u [] = return $ u
-- app u c = do  
--          case (u,c) of
--             (VApp u2 c2,_) -> app u2 (c2 ++ c)
--             (VLam x env e,(v:vl))  -> do v' <- whnf (update env x v) e
--                                          app v' vl
--             (VDef n,_) -> appDef n c
--             (VUp v (VPi dec x av rho b), w:wl) -> do
-- {- 
-- -- ALT: erased functions are not applied to their argument!
--               v1 <- if erased dec then return v else app v [w]  -- eta-expand w ??
-- -}
--               v1 <- app v [w]  -- eta-expand w ??
--               bv <- whnf (update rho x w) b
--               v2 <- up v1 bv
--               app v2 wl
-- {-
-- -- ALT: VIrr consumes applications
--             (VIrr,_) -> return VIrr
--  -}
--             (VIrr,_) -> fail $ "app internal error: " ++ show (VApp u c)
--             _ -> return $ VApp u c
 

-- unroll a corecursive definition one time (until constructor appears)
force' :: Bool -> Val -> TypeCheck (Bool, Val)
force' b (VSing v tv) = do  -- for singleton types, force type!
  (b',tv') <- force' b tv
  return (b', VSing v tv')
force' b (VUp v tv) = up True v tv >>= \ v' -> return (True, v')  -- force eta expansion
force' b (VClos rho e) = do
  v <- whnf rho e
  force' b v
force' b v@(VDef (DefId Fun n)) = failValInv v
{-
 --trace ("force " ++ show v) $
    do sig <- gets signature
       case lookupSig n sig of
         (FunSig CoInd t cl True) -> do m <- matchClauses [] cl []
                                        case m of 
                                          Just v' -> force v'
                                          Nothing -> return v
         _ -> return v
-}
force' b v@(VApp (VDef (DefId Fun n)) vl) = --trace ("force " ++ show v) $
    do sige <- lookupSymb n 
       case sige of
         (FunSig isCo t ar cl True) -> traceMatch ("forcing " ++ show v) $
            do m <- matchClauses emptyEnv cl vl
               case m of 
                 Just v' -> traceMatch ("forcing " ++ n ++ " succeeded") $
                   force' True v'
                 Nothing -> traceMatch ("forcing " ++ n ++ " failed") $
                   return (b, v)
         _ -> return (b, v)
force' b v = return (b, v)

force :: Val -> TypeCheck Val
force v = -- trace ("forcing " ++ show v) $
  liftM snd $ force' False v

-- apply a recursive function
-- corecursive ones are not expanded even if the arity is exceeded 
-- this is because a coinductive type needs to be destructed by pattern matching
appDef :: Name -> [Val] -> TypeCheck Val
appDef n vl = --trace ("appDef " ++ n) $ 
    do
      sige <- lookupSymb n
      case sige of
        FunSig { isCo = Ind, arity = ar, clauses = cl, isTypeChecked = True } | length vl >= ar -> do 
           m <- matchClauses emptyEnv cl vl
           case m of
              Nothing -> return $ VApp (VDef (DefId Fun n)) vl
              Just v2 -> return v2
        _ -> return $ VApp (VDef (DefId Fun n)) vl   

-- reflection and reification  ---------------------------------------

-- up force v tv
-- force==True also expands at coinductive type
up :: Bool -> Val -> TVal -> TypeCheck Val
up f (VUp v tv') tv = up f v tv
up f v tv@(VPi dec x va rho b) = return $ VUp v tv
up f _ (VSing v vt) = up f v vt
up f v (VDef d) = failValInv (VDef d) -- upData v d []
up f v (VApp (VDef (DefId Dat d)) vl) = upData f v d vl
up f v _ = return $ v

{- Most of the code to eta expand on data types is in 
   TypeChecker.hs "typeCheckDeclaration"

 Currently, eta expansion only happens at data *types* with exactly
one constructor.  In a first step, this will be extended to
non-recursive pattern inductive families.

The strategy is: match type value with result type for all the constructors
0. if there are no matches, eta expand to * (VIrr)
1. if there is exactly one match, eta expand accordingly using destructors
2. if there are more matches, do not eta-expand

up{Vec A (suc n)} x = vcons A n (head A n x) (tail A n x)

up{Vec Bool (suc zero)} x 
  = vcons Bool zero (head Bool zero x) (tail Bool zero x)
 
For vcons 
- the patterns of  Vec : (A : Set) -> Nat -> Set  are  [A,suc n]
- matching  Bool,suc zero  against  A,suc n  yields A=Bool,n=zero
- this means we can eta expand to vcons
- go through the fields of vcons
  - if Index use value obtained by matching
  - if Field destr, use  destr <all pars> <all indices> x
   
-}

-- matchingConstructors is for use in checkPattern
-- matchingConstructors (D vs)  returns all the constructors 
-- each as tuple (ci,rho)
-- of family D whose target matches (D vs) under substitution rho
matchingConstructors :: Val -> TypeCheck (Maybe [(ConstructorInfo,Env)])
matchingConstructors v@(VDef d) = failValInv v -- matchingConstructors' d []
matchingConstructors (VApp (VDef (DefId Dat d)) vl) = matchingConstructors' d vl >>= return . Just
matchingConstructors v = return Nothing
-- fail $ "matchingConstructors: not a data type: " ++ show v -- return []

matchingConstructors' :: Name -> [Val] -> TypeCheck [(ConstructorInfo,Env)]
matchingConstructors' n vl = do
  sige <- lookupSymb n
  case sige of
    (DataSig {symbTyp = dv, constructors = cs}) -> -- if (null cs) then ret [] else do -- no constructor
      matchingConstructors'' True vl dv cs

-- matchingConstructors'' 
-- Arguments:
--   symm     symmetric match
--   vl       arguments to D (instance of D)
--   dv       complete type value of D
--   cs       constructors
-- Returns a list [(ci,rho)] of matching constructors together with the 
--   environments which are solutions for the free variables in the constr.type
-- this is also for use in upData
matchingConstructors'' :: Bool -> [Val] -> Val -> [ConstructorInfo] -> TypeCheck [(ConstructorInfo,Env)]
matchingConstructors'' symm vl dv cs = do
      vl <- mapM whnfClos vl
      cenvs <- mapM (\ ci -> let ps = snd (cPatFam ci) -- list of patterns ps where D ps is the constructor target
                             in liftM (fmap ((,) ci)) $ nonLinMatchList symm emptyEnv ps vl dv) cs
      return $ compressMaybes cenvs

-- eta expand  v  at data type  n vl
upData :: Bool -> Val -> Name -> [Val] -> TypeCheck Val
upData force v n vl = -- trace ("upData " ++ show v ++ " at " ++ n ++ show vl) $ 
 do
  let ret v' = traceEta ("Eta-expanding: " ++ show v ++ " --> " ++ show v' ++ " at type " ++ n ++ show vl) $ return v'
  sige <- lookupSymb n
  case sige of
    (DataSig {symbTyp = dv, numPars = npars, isCo = co, constructors = cs, etaExpand = True}) -> if (null cs) then ret VIrr else do -- no constructor: empty type
      let (pars, inds) = splitAt npars vl
      -- for each constructor, match its core against the type
      -- produces a list of maybe (c.info, environment)
      cenvs <- matchingConstructors'' False vl dv cs      
      -- traceM $ "Matching constructors: " ++ show cenvs
      case cenvs of
        -- exactly one matching constructor: can eta expand
        [(ci,env)] -> if not (cEtaExp ci) then return v else 
         if (co==CoInd && not force) then return $ VUp v (VApp (VDef $ DefId Dat n) vl) else do
          -- get list of index values from environment
          let fis = cFields ci
          let indices = filter (\ fi -> fClass fi == Index) fis
          let indvs = map (\ fi -> lookupPure env (fName fi)) indices 
          let piv = pars ++ indvs ++ [v]
          -- for each argument of constructor, get value
          let arg (FieldInfo { fName = x, fClass = Index }) = 
                lookupEnv env x 
              arg (FieldInfo { fName = d, fClass = Field _ }) = do
                -- lookup type sig  t  of destructor  d
                FunSig { symbTyp = t } <- lookupSymb d 
                -- pi-apply destructor type to parameters, indices and value v
                t' <- piApps t piv
                -- recursively eta expand  (d <pars> v)
                -- WAS: up (VDef (DefId Fun d) `VApp` piv) t'
                up False (VDef (DefId Fun d) `VApp` piv) t' -- now: LAZY
          vs <- mapM arg fis 
          v' <- foldM app (vCon co (cName ci)) (pars ++ vs) 
          ret v'
        -- more or less than one matching constructors: cannot eta expand
        l -> -- trace ("Eta: " ++ show (length l) ++ " patterns match at type " ++ show n ++ show vl) $ 
               return v
    _ -> return v


{-
      let matchC (c, ps, ds) = 
            do menv <- nonLinMatchList [] ps inds dv 
               case menv of
                 Nothing -> return False
                 Just env -> do
                   let grps = groupBy (\ (x,_) (y,_) -> x == y) env
                   -- TODO: now compare elements in the group
                   -- NEED types for equality check
                   -- trivial if groups are singletons
                   return $ all (\ l -> length l <= 1) grps
      cs' <- filterM matchC cs 
      case cs' of
        [] -> return $ VIrr
        [(c,_,ds)] ->  do
          let parsv = pars ++ [v]
          let aux d = do
               -- lookup type sig  t  of destructor  d
               let FunSig { symbTyp = t } = lookupSig d sig
               -- pi-apply destructor type to parameters and value v
               t' <- piApps t parsv
               -- recursively eta expand  (d <pars> v)
               up (VDef d `VApp` parsv) t'
          vs <- mapM aux ds 
          app (VCon co c) (pars ++ vs) 
        _ -> return v
    _ -> return v
-}
   
{- 
refl : [A : Set] -> [a : A] -> Id A a a
up{Id T t t'} x 
  Id T t t' =?= Id A a a  --> A = T, a = t, a = t'
-}
      
{- OLD CODE FOR NON-DEPENDENT RECORDS ONLY
    -- erase if n is a empty type
    (DataSig {constructors = []}) -> return $ VIrr
    -- eta expand v if n is a tuple type
    (DataSig {isCo = co, constructors = [c], destructors = Just ds}) -> do
       let vlv = vl ++ [v]
       let aux d = do -- lookup type sig  t  of destructor  d
                      let FunSig { symbTyp = t } = lookupSig d sig
                      -- pi-apply destructor type to parameters and value v
                      t' <- piApps t vlv
                      -- recursively eta expand  (d <pars> v)
                      up (VDef d `VApp` vlv) t'
       vs <- mapM aux ds 
       app (VCon co c) (vl ++ vs) -- (map (\d -> VDef d `VApp` (vl ++ [v])) ds)
    _ -> return v
END OLD CODE -}

-- pattern matching ---------------------------------------------------

matchClauses :: Env -> [Clause] -> [Val] -> TypeCheck (Maybe Val)
matchClauses env cl vl0 = do
  vl <- mapM reduce vl0  -- REWRITE before matching (2010-07-12 dysfunctional because of lazy?)
  loop cl vl
    where loop [] vl = return Nothing
          loop (Clause pl Nothing : cl2) vl = loop cl2 vl -- no need to try absurd clauses
          loop (Clause pl (Just rhs) : cl2) vl = 
              do x <- matchClause env pl rhs vl 
                 case x of
                   Nothing -> loop cl2 vl
                   Just v -> return $ Just v

matchClause :: Env -> [Pattern] -> Expr -> [Val] -> TypeCheck (Maybe Val)
matchClause env [] rhs vl = do v <- whnf env rhs
                               v2 <- foldM app v vl
                               return $ Just v2
matchClause env (p:pl) rhs (v:vl) =
  do m <- match env p v 
     case m of
       Just env' ->
         matchClause env' pl rhs vl
       Nothing -> return Nothing
-- too few arguments
matchClause _ _ _ [] = return Nothing

match :: Env -> Pattern -> Val -> TypeCheck (Maybe Env)
match env p v0 = --trace (show env ++ show v0) $ 
  do 
    -- force against constructor pattern
    v <- case p of 
           (ConP _ _ _) -> traceMatch ("matching pattern " ++ show p) $ force v0
           _ -> return v0
    case (p,v) of
--      (ErasedP _,_) -> return $ Just env  -- TOO BAD, DOES NOT WORK (eta!)
      (ErasedP p,_) -> match env p v
      (DotP _,   _) -> return $ Just env
      (VarP x,   _) -> return $ Just (update env x v)
      (SizeP _ x,_) -> return $ Just (update env x v)
      (ConP _ x [],VDef (DefId (Con _) y)) -> failValInv v -- | x == y -> return $ Just env
      (ConP _ x pl,VApp (VDef (DefId (Con _) y)) vl) | x == y -> matchList env pl vl
      (SuccP p',v) -> do
         v <- whnfClos v
         case predSize v of
            Nothing -> return Nothing  
            Just v' -> match env p' v'
--      (SuccP p',VInfty) -> match env p' VInfty 
--      (SuccP p',VSucc v') -> match env p' v'
      _ -> return Nothing

matchList :: Env -> [Pattern] -> [Val] -> TypeCheck (Maybe Env)
matchList env [] [] = return $ Just env
matchList env (p:pl) (v:vl) = do m <- match env p v 
                                 case m of 
                                   Just env' -> matchList env' pl vl
                                   Nothing -> return Nothing
matchList env pl vl = fail $ "matchList internal error: inequal length while to match patterns " ++ show pl ++ " against values " ++ show vl

-- typed non-linear matching -----------------------------------------

-- nonLinMatch True  allows also instantiation in v0
-- this is useful for finding all matching constructors
-- for an erased argument in checkPattern
nonLinMatch :: Bool -> Env -> Pattern -> Val -> TVal -> TypeCheck (Maybe Env)
nonLinMatch symm env p v0 tv = do 
    -- force against constructor pattern
      v <- case p of 
             (ConP _ _ _) -> force v0
             _ -> return v0
      case (p,v) of
        (ErasedP _,_) -> return $ Just env
        (DotP _,_) -> return $ Just env
        (_, VGen _) | symm -> return $ Just env -- no check in case of non-lin!
        (VarP x,_) -> case find (\ (y,v') -> x == y) (envMap env) of
                        Nothing -> return $ Just (update env x v)
                        Just (y,v') -> do 
                          b <- eqValBool tv v v' 
                          if b then return $ Just env else return Nothing
        (SizeP _ x,_) -> {- CUT-AND-PASTE-CODE, BAD! -} 
                      case find (\ (y,v') -> x == y) (envMap env) of
                        Nothing -> return $ Just (update env x v)
                        Just (y,v') -> do 
                          b <- eqValBool tv v v' 
                          if b then return $ Just env else return Nothing
        (ConP _ c [],VDef (DefId (Con _) c')) -> failValInv v -- | c == c' -> return $ Just env
        (ConP _ c pl,VApp (VDef (DefId (Con _) c')) vl) | c == c' -> do
           sy <- lookupSymb c
           nonLinMatchList symm env pl vl (symbTyp sy)
        (SuccP p',v) -> do
           v <- whnfClos v
           case predSize v of
              Nothing -> return Nothing  
              Just v' -> nonLinMatch symm env p' v' tv
--        (SuccP p',VInfty) -> nonLinMatch symm env p' VInfty tv
--        (SuccP p',VSucc v') -> nonLinMatch symm env p' v' tv
        _ -> return Nothing

-- nonLinMatchList symm env ps vs tv 
-- typed non-linear matching of patterns ps against values vs at type tv
--   env   is the accumulator for the solution of the matching
nonLinMatchList :: Bool -> Env -> [Pattern] -> [Val] -> TVal -> TypeCheck (Maybe Env)
nonLinMatchList symm env [] [] tv = return $ Just env
nonLinMatchList symm env (p:pl) (v:vl) tv =
  case tv of
    VPi dec x va rho b -> do 
      m <- nonLinMatch symm env p v va 
      case m of 
         Just env' -> do
            vb <- whnf (update rho x v) b
            nonLinMatchList symm env' pl vl vb
         Nothing -> return Nothing
    _ -> fail $ "nonLinMatchList: cannot match in absence of pi-type"
nonLinMatchList _ _ _ _ _ = return Nothing



-- heterogeneous typed equality and subtyping ------------------------

eqValBool :: TVal -> Val -> Val -> TypeCheck Bool
eqValBool tv v v' = errorToBool $ eqVal tv v v'
-- eqValBool tv v v' = (eqVal tv v v' >> return True) `catchError` (\ _ -> return False)

eqVal :: TVal -> Val -> Val -> TypeCheck ()
eqVal tv = leqVal' N Mixed (Just (tv,tv))


-- force history
data Force = N | L | R -- not yet, left , right  
    deriving (Eq,Show)

class Switchable a where
  switch :: a -> a

instance Switchable Force where
  switch L = R
  switch R = L
  switch N = N

instance Switchable Pos where
  switch = polNeg

instance Switchable (a,a) where
  switch (a,b) = (b,a)

instance Switchable a => Switchable (Maybe a) where
  switch = fmap switch

{- 
-- WONTFIX: FOR THE FOLLOWING TO BE SOUND, ONE NEEDS COERCIVE SUBTYPING!
-- the problem is that after extraction, erased arguments are gone!
-- a function which does not use its argument can be used as just a function
-- [A] -> A <= A -> A
-- A <= [A]
leqDec :: Pos -> Dec -> Dec -> Bool
leqDec SPos  dec1 dec2 = erased dec2 || not (erased dec1)
leqDec Neg   dec1 dec2 = erased dec1 || not (erased dec2)
leqDec Mixed dec1 dec2 = erased dec1 == erased dec2 
-}

-- subtyping for erasure disabled
-- but subtyping for polarities!
leqDec :: Pos -> Dec -> Dec -> Bool
leqDec p dec1 dec2 = erased dec1 == erased dec2 
  && relPol p leqPol (polarity dec1) (polarity dec2)

-- subtyping ---------------------------------------------------------

subtype :: Val -> Val -> TypeCheck ()
subtype v1 v2 = -- enter ("subtype " ++ show v1 ++ "  <=  " ++ show v2) $
  leqVal' N Pos Nothing v1 v2

-- Pos ::= Pos | Neg | Mixed
leqVal :: Pos -> TVal -> Val -> Val -> TypeCheck ()
leqVal p tv = leqVal' N p (Just (tv, tv))

type MT12 = Maybe (TVal, TVal)

-- if m12 = Nothing, we are checking subtyping, otherwise we are
-- comparing objects or higher-kinded types
leqVal' :: Force -> Pos -> MT12 -> Val -> Val -> TypeCheck ()
leqVal' f p mt12 u1' u2' = do 
 l <- getLen 
 ren <- getRen
 enter (case mt12 of
  Nothing -> ("leqVal' (subtyping) " ++ show  (Map.toList $ ren) ++ " |- " ++ show u1' ++ "  <=" ++ show p ++ "  " ++ show u2')
  Just (tv1,tv2) -> ("leqVal' " ++ show  (Map.toList $ ren) ++ " |- " ++ show u1' ++ "  :  " ++ show tv1 ++ "\n  <=" ++ show p ++ "  " ++ show u2' ++ "  :  " ++ show tv2)) $
{-   do
    ce <- ask
    trace  (("rewrites: " +?+ show (rewrites ce)) ++ "  leqVal': " ++ show ce ++ "\n |- " ++ show u1' ++ "\n  <=" ++ show p ++ "  " ++ show u2') $
-}  
    case mt12 of
      Just (VSing v1 tv1', tv2) -> leqVal' f p (Just (tv1', tv2)) v1 u2'
      Just (tv1, VSing v2 tv2') -> leqVal' f p (Just (tv1, tv2')) u1' v2
      Just (VSort Size, VSort Size) -> leqSize p u1' u2'

{-  functions are compared pointwise

   Gamma, p(x:A) |- t x : B  <=  Gamma', p'(x:A') |- t' x : B'
   ----------------------------------------------------------
   Gamma |- t : p(x:A) -> B  <=  Gamma' |- t' : p'(x:A') -> B'
-}
      Just (VPi dec1 x1 av1 env1 b1, VPi dec2 x2 av2 env2 b2)  -> 
         new2 x1 (Domain av1 dec1, Domain av2 dec2) $ \ (xv1, xv2) -> do
            u1' <- app u1' xv1
            u2' <- app u2' xv2
            tv1' <- whnf (update env1 x1 xv1) b1
            tv2' <- whnf (update env2 x2 xv2) b2
            leqVal' f p (Just (tv1', tv2')) u1' u2'           

--            Nothing -> case (u1,u2) of -- subtyping of sets

      -- structural subtyping (not directed by types) 

      _ -> do 
       u1 <- whnfClos u1' >>= reduce
       u2 <- whnfClos u2' >>= reduce
       (f1,u1f) <- force' False u1
       (f2,u2f) <- force' False u2
       case (f1,f2) of -- (u1f /= u1,u2f /= u2) of
         (True,False) | f /= R -> -- only unroll one side  
            enter ("forcing LHS") $
                      leqVal' L p mt12 u1f u2
         (False,True) | f /= L ->
            enter ("forcing RHS") $
                      leqVal' R p mt12 u1 u2f    
 
         -- subtyping directed by type tv
 
         _ -> case (u1,u2) of

{-
  C = C'  (proper: C' entails C, but I do not want to implement entailment)     
  Gamma, C |- A  <=  Gamma', C' |- A'
  -----------------------------------------
  Gamma |- C ==> A  <=  Gamma' |- C' ==> A'
-}
              (VGuard beta1 bv1, VGuard beta2 bv2) -> do
                 entailsGuard (switch p) beta1 beta2
                 leqVal' f p Nothing bv1 bv2
 {-
  p' <= p   
  Gamma' |- A' <= Gamma |- A
  Gamma, p(x:A) |- B <= Gamma', p'(x:A') |- B'
  ---------------------------------------------------------
  Gamma |- p(x:A) -> B : s <= Gamma' |- p'(x:A') -> B' : s'
-}
              (VPi dec1 x1 av1 env1 b1, VPi dec2 x2 av2 env2 b2) ->
                 if not $ leqDec p dec2 dec1 then
                    throwErrorMsg $ "subtyping " ++ show u1 ++ " <=" ++ show p ++ " " ++ show u2 ++ " failed"
                  else do     
                    leqVal' (switch f) (switch p) Nothing av1 av2
                    new2 x1 (Domain av1 dec1, Domain av2 dec2) $ \ (xv1, xv2) -> do
                      bv1 <- whnf (update env1 x1 xv1) b1
                      bv2 <- whnf (update env2 x2 xv2) b2
                      enter ("comparing codomain " ++ show b1 ++ " with " ++ show b2) $
                        leqVal' f p Nothing bv1 bv2

              (VSing v1 av1, VSing v2 av2) -> do
                  leqVal' f p Nothing av1 av2
                  leqVal' N Mixed (Just (av1,av2)) v1 v2  -- compare for eq.     

              -- unresolved eta-expansions (e.g. at coinductive type)
{-
              (VUp v1 av1, u2) -> leqVal' f p mt12 v1 u2 
              (u1, VUp v2 av2) -> leqVal' f p mt12 u1 v2
-}
              (VUp v1 av1, VUp v2 av2) -> do
                  -- leqVal' f p Nothing av1 av2      -- do not compare types
                  leqVal' f p (Just (av1,av2)) v1 v2  -- OR: Just(tv1,tv2) ?
              (VApp v1 vl1,VApp v2 vl2) -> do
                           leqApp f p v1 vl1 v2 vl2
              (VCase v env cl, v2) -> leqCases (switch f) (switch p) (switch mt12) v2 v env cl
              (v1, VCase v env cl) -> leqCases f p mt12 v1 v env cl
              (VSing v1 av1, av2)  -> leqVal' f p Nothing av1 av2  -- subtyping ax 
              (VSort s1, VSort s2) -> leqSort p s1 s2
              (a1,a2) | a1 == a2 -> return ()
              _ -> leqApp f p u1 [] u2 []
                   

-- leqCases f p tv v1 v env cl 
-- checks whether  v1 <=p (VCase v env cl) : tv
leqCases :: Force -> Pos -> MT12 -> Val -> Val -> Env -> [Clause] -> TypeCheck ()
leqCases f pol mt12 v1 v env cl = do
  vcase <- evalCase v env cl
  case vcase of 
    (VCase v env cl) -> mapM_ (leqCase f pol mt12 v1 v env) cl
    v2 -> leqVal' f pol mt12 v1 v2

-- absurd cases need not be checked
leqCase :: Force -> Pos -> MT12 -> Val -> Val -> Env -> Clause -> TypeCheck ()
leqCase f pol mt12 v1 v env (Clause [p] Nothing) = return ()
leqCase f pol mt12 v1 v env (Clause [p] (Just e)) = enter ("leqCase " ++ show v ++ "  -->  " ++ show p ++ "  |-  " ++ show v1 ++ "  <=" ++ show pol ++ "  " ++ show e ++ " in " ++ show env) $ do    -- ++ "  :  " ++ show mt12) $
-- the dot patterns inside p are only valid in environment env
  let tv = case mt12 of 
             Nothing -> VSort Type
             Just (tv1,tv2) -> tv2
  addPattern (dontCare `arrow` tv) p env $ \ _ pv env' ->
    addRewrite (Rewrite v pv) [tv,v1] $ \ [tv',v1'] -> do
      v2  <- whnf (appendEnv env' env) e
      v2' <- reval v2
      leqVal' f pol (fmap (\(tv1,tv2) -> (tv1,tv')) mt12) v1' v2'
   where dontCare = VSort Tm -- bogus by intention -- (error "internal error in leqCase: unknown type")

-- compare spines (see rule Al-App-Ne, Abel, MSCS 08)
leqVals' :: Force -> Pos -> (TVal, TVal) -> [Val] -> [Val] -> TypeCheck (TVal, TVal)
leqVals' f q tv12 [] [] = return $ tv12
leqVals' f q (VPi dec1 x1 av1 env1 b1, VPi dec2 x2 av2 env2 b2) (w1:vs1) (w2:vs2) | dec1 == dec2 = do
  let p = polarity dec1
  v1 <- whnfClos w1
  v2 <- whnfClos w2
  when (not (erased dec1)) $ 
    applyDec dec1 $ leqVal' f (polComp p q) (Just (av1,av2)) v1 v2 
  tv1 <- whnf (update env1 x1 v1) b1
  tv2 <- whnf (update env2 x2 v2) b2
  leqVals' f q (tv1,tv2) vs1 vs2
leqVals' f q tv12 vl1 vl2 = throwErrorMsg 
  "leqVals': not (compatible) function types or mismatch number of arguments"

{-
leqNe :: Force -> Val -> Val -> TypeCheck TVal
leqNe f v1 v2 = --trace ("leqNe " ++ show v1 ++ "<=" ++ show v2) $ 
  do case (v1,v2) of
      (VGen k1, VGen k2) -> if k1 == k2 then do 
                                 dom <- lookupGem k1
                                 return $ typ dom
                               else throwErrorMsg $ "gen mismatch "  ++ show k1 ++ " " ++ show k2 
-}
     
leqApp :: Force -> Pos -> Val -> [Val] -> Val -> [Val] -> TypeCheck ()
leqApp f pol v1 w1 v2 w2 = {- trace ("leqApp: " -- ++ show delta ++ " |- " 
                                  ++ show v1 ++ show w1 ++ " <=" ++ show pol ++ " " ++ show v2 ++ show w2) $ -}
  do case (v1,v2) of
      (VGen k1, VGen k2) -> 
        if k1 /= k2 
          then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2 
          else do (tv1,tv2) <- lookupGenType2 k1
                  leqVals' f pol (tv1,tv2) w1 w2
                  return ()
{-
      (VCon _ n, VCon _ m) -> 
        if n /= m 
         then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2  
         else do 
          sige <- lookupSymb n
          case sige of
            (ConSig tv) -> -- constructor
               leqVals' f tv (repeat Mixed) w1 w2 >> return ()
-}
      (VDef n, VDef m) -> 
        if (name n) /= (name m) 
         then throwErrorMsg $ 
            "leqApp: head mismatch "  ++ show v1 ++ " != " ++ show v2  
         else do 
          entry <- lookupSymb (name n)  
          leqVals' f pol (symbTyp entry, symbTyp entry) w1 w2 
          return ()
{-
          sig <- gets signature
          case lookupSig (name n) sig of
            (DataSig{ numPars = p, positivity = pos, isSized = s, isCo = co, symbTyp = tv }) -> -- data type 
               let positivitySizeIndex = if s /= Sized then Mixed else 
                                           if co == Ind then Pos else Neg
                   pos' = -- trace ("leqApp:  posOrig = " ++ show (pos ++ [positivitySizeIndex])) $
                     map (polComp pol) (pos ++ positivitySizeIndex : repeat Mixed) -- the polComp will replace all SPos by Pos 
               in leqVals' f tv pos' w1 w2
                    >> return ()

-- otherwise, we are dealing with a (co) recursive function or a constructor
            entry -> leqVals' f (symbTyp entry) (repeat Mixed) w1 w2 >> return ()
-}

{-
            (ConSig tv) -> -- constructor
               leqVals' f tv (repeat Mixed) w1 w2 >> return ()
            (FunSig co tv _ _) ->  
               leqVals' f tv (repeat Mixed) w1 w2 >> return ()
-}
                                                 
      _ -> throwErrorMsg $ "leqApp: " ++ show v1 ++ show w1 ++ " !<= " ++ show v2 ++ show w2
{-
             do leqVal delta v1 v2 
                eqVals delta w1 w2
 -}

-- comparing sorts and sizes -----------------------------------------

leqSort :: Pos -> Sort Val -> Sort Val -> TypeCheck ()
leqSort p = relPolM p leqSort' 
{-
leqSort Mixed s1 s2 = leqSort' s1 s2 >> leqSort' s2 s1
leqSort Neg s1 s2 = leqSort' s2 s1
leqSort Pos s1 s2 = leqSort' s1 s2
-}

leqSort' :: Sort Val -> Sort Val -> TypeCheck ()
leqSort' s1 s2 = do
  let err = "universe test " ++ show s1 ++ " <= " ++ show s2 ++ " failed"
  case (s1,s2) of 
     (_,Type) -> return ()
     (Type,_) -> fail err
     (TSize,TSize) -> return ()
     (Size,Size) -> return ()
     (Set v1,Set v2)   -> leqSize Pos v1 v2
     (CoSet VInfty,Set v) -> return ()
     (Set VZero, CoSet{}) -> return () 
     (CoSet v1, CoSet v2) -> leqSize Neg v1 v2
     _ -> recoverFail err

minSize :: Val -> Val -> Maybe Val
minSize v1 v2 = 
  case (v1,v2) of
    (VZero,_)  -> return VZero
    (_,VZero)  -> return VZero
    (VInfty,_) -> return v2
    (_,VInfty) -> return v1
    (VMax vs,_) -> maxMins $ map (\ v -> minSize v v2) vs
    (_,VMax vs) -> maxMins $ map (\ v -> minSize v1 v) vs
    (VSucc v1', VSucc v2') -> fmap succSize $ minSize v1' v2'
    (VGen i, VGen j) -> if i == j then return $ VGen i else Nothing
    (VSucc v1', VGen j) -> minSize v1' v2
    (VGen i, VSucc v2') -> minSize v1 v2'

maxMins :: [Maybe Val] -> Maybe Val
maxMins mvs = case compressMaybes mvs of
                     [] -> Nothing
                     vs' -> return $ maxSize vs' 

-- substaging on size values
leqSize :: Pos -> Val -> Val -> TypeCheck ()
leqSize pol v1 v2 = -- trace ("leqSize " ++ show v1 ++ " <=" ++ show pol ++ " " ++ show v2) $ 
    do case (v1,v2) of
         (VSucc v1,VSucc v2) -> leqSize pol v1 v2
         (VGen i1,VGen i2) -> do 
           d <- getSizeDiff i1 i2
           case d of
             Nothing -> recoverFail $ "leqSize: head mismatch: " ++ show v1 ++ " !<= " ++ show v2
             Just k -> case (pol,k) of
               (Mixed, 0) -> return ()
               (Pos, _) | k >= 0 -> return ()
               (Neg, _) | k <= 0 -> return ()
               _ ->  recoverFail $ "leqSize: " ++ show v1 ++ " !<=" ++ show pol ++ " " ++ show v2 ++ " failed"
{-           
           if v1 == v2 then return ()
           else throwErrorMsg $ "leqSize: head mismatch: " ++ show v1 ++ " !<= " ++ show v2 
-}
         (VInfty,VInfty) -> return ()
         (VApp h1 tl1,VApp h2 tl2) -> leqApp N pol h1 tl1 h2 tl2
         _ -> case pol of
                Pos   -> leqSize' v1 v2
                Mixed -> leqSize' v1 v2 >> leqSize' v2 v1
                Neg   -> leqSize' v2 v1

leqSize' ::  Val -> Val -> TypeCheck ()  
leqSize' v1 v2 = --trace ("leqSize' " ++ show v1 ++ show v2) $ 
    do let err = "leqSize': " ++ show v1 ++ " <= " ++ show v2 ++ " failed"
       case (v1,v2) of
         (VZero,_)   -> return ()
         (VSucc{}, VZero) -> fail err
         (VInfty, VZero) -> fail err
         (VGen{}, VZero) -> fail err
         (VMax vs,_) -> mapM_ (\ v -> leqSize' v v2) vs -- all v in vs <= v2
         (_,VMax _)  -> addLeq v1 v2 -- this produces a disjunction 
         (_,VInfty) -> return ()
         (VMeta{},VZero) -> addLeq v1 v2
{-
         (0,VMeta i n', VMeta j m') -> 
           let (n,m) = if bal <= 0 then (n', m' - bal) else (n' + bal, m') in
-}
         (VMeta i n, VMeta j m) -> 
               addLeq   (VMeta i (n - min n m))
                        (VMeta j (m - min n m)) 
         (VMeta i n, VSucc v2) | n > 0 -> leqSize' (VMeta i (n-1)) v2
         (VMeta i n, v2)  -> addLeq v1 v2
         (VSucc v1, VMeta i n) | n > 0 -> leqSize' v1 (VMeta i (n-1))
         (v1,VMeta i n) -> addLeq v1 v2
         _ -> leqSize'' 0 v1 v2

-- leqSize'' bal v v'  checks whether  Succ^bal v <= v'
-- invariant: bal is zero in cases for VMax and VMeta
leqSize'' ::  Int -> Val -> Val -> TypeCheck ()  
leqSize'' bal v1 v2 = -- trace ("leqSize'' " ++ show v1 ++ " + " ++ show bal ++ " <= " ++ show v2) $ 
    do case (v1,v2) of
         (VZero,_) | bal <= 0 -> return ()
         (VSucc v1, v2) -> leqSize'' (bal + 1) v1 v2
         (v1, VSucc v2) -> leqSize'' (bal - 1) v1 v2
         (VGen i1,VGen i2) -> do 
           d <- getSizeDiff i1 i2
           case d of
             Nothing -> recoverFail $ "leqSize: head mismatch: " ++ show v1 ++ " !<= " ++ show v2
             Just k -> if k >= bal && k >= 0 then return () else recoverFail $ "leqSize: " ++ show v1 ++ " + " ++ show bal ++ " !<= " ++ show v2 ++ " failed" 
         _ -> recoverFail $ "leqSize: " ++ show v1 ++ " + " ++ show bal ++ " !<= " ++ show v2

{-
leqSize' :: Val -> Val -> TypeCheck ()  
leqSize' v1 v2 = --trace ("leqSize' " ++ show v1 ++ show v2) $ 
    do case (v1,v2) of
         (VMax vs,_) -> mapM_ (\ v -> leqSize' v v2) vs -- all v in vs <= v2
         (_,VMax _)  -> addLeq v1 v2 -- this produces a disjunction 
         (VSucc v1,VSucc v2) -> leqSize' v1 v2
         (VGen v1,VGen v2) -> do 
           d <- getSizeDiff v1 v2
           case d of
             Nothing -> throwErrorMsg $ "leqSize: head mismatch: " ++ show v1 ++ " !<= " ++ show v2
             Just k -> if k >= 0 then return () else throwErrorMsg $ "leqSize: " ++ show v1 ++ " !<= " ++ show v2 ++ " failed" 
         (_,VInfty) -> return ()
         (VMeta i n, VSucc v2) | n > 0 -> leqSize' (VMeta i (n-1)) v2
         (VMeta i n, VMeta j m) -> addLeq (VMeta i (n - min n m))
                                          (VMeta j (m - min n m)) 
         (VMeta i n, v2) -> addLeq v1 v2
         (VSucc v1, VMeta i n) | n > 0 -> leqSize' v1 (VMeta i (n-1))
         (v1,VMeta i n) -> addLeq v1 v2
         (v1,VSucc v2) -> leqSize' v1 v2
         _ -> throwErrorMsg $ "leqSize: " ++ show v1 ++ " !<= " ++ show v2
-}

-- compare lexicographically
-- precondition: same length
ltMeasure :: Measure Val -> Measure Val -> TypeCheck ()
ltMeasure  (Measure mu1) (Measure mu2) = 
  -- enter ("checking " ++ show mu1 ++ " < " ++ show mu2) $ 
    lexSizes True mu1 mu2
{-
leqMeasure :: Pos -> Measure Val -> Measure Val -> TypeCheck ()
leqMeasure Mixed (Measure mu1) (Measure mu2) = do
  zipWithM (leqSize Mixed) mu1 mu2
  return ()
leqMeasure Pos (Measure mu1) (Measure mu2) = lexSizes mu1 mu2
leqMeasure Neg (Measure mu1) (Measure mu2) = lexSizes mu2 mu1
-}

lexSizes :: Bool -> [Val] -> [Val] -> TypeCheck ()
lexSizes True [] [] = recoverFail $ "lexSizes: no descent detected"
lexSizes False [] [] = return ()
lexSizes lt (a1:mu1) (a2:mu2) = do
    r <- compareSize a1 a2 
    case r of
      LT -> return ()         
      EQ -> lexSizes lt mu1 mu2
      GT -> recoverFail $ "lexSizes: expected " ++ show a1 ++ " <= " ++ show a2

-- TODO: reprogram leqSize in terms of a proper compareSize
compareSize :: Val -> Val -> TypeCheck Ordering
compareSize a1 a2 = do
  le <- errorToBool $ leqSize Pos a1 a2 
  ge <- errorToBool $ leqSize Pos a2 a1
  case (le,ge) of
    (True,False) -> return LT
    (True,True)  -> return EQ
    (False,True) -> return GT
    (False,False) -> fail $ "compareSize (" ++ show a1 ++ ", " ++ show a2 ++ "): sizes incomparable"

{-
entailsGuard :: Pos -> Bound Val -> Bound Val -> TypeCheck ()
entailsGuard pol beta1@(Bound (Measure mu1) (Measure mu1')) beta2@(Bound (Measure mu2) (Measure mu2')) = enterTrace ("entailsGuard: " ++ show beta1 ++ " " ++ show pol ++ "==> " ++ show beta2) $ do
  zipWithM (leqSize Mixed) mu1 mu2
  case pol of
    Mixed -> zipWithM (leqSize Mixed) mu1' mu2' >> return ()
    Pos   -> lexSizes False mu1' mu2' >> return () -- not strictly smaller
    Neg   -> lexSizes False mu2' mu1' >> return ()
-}  

entailsGuard :: Pos -> Bound Val -> Bound Val -> TypeCheck ()
entailsGuard pol beta1@(Bound (Measure mu1) (Measure mu1')) beta2@(Bound (Measure mu2) (Measure mu2')) = enterTrace ("entailsGuard: " ++ show beta1 ++ " " ++ show pol ++ "==> " ++ show beta2) $ do
  case pol of
    Mixed -> do
      zipWithM (leqSize Mixed) mu1  mu2
      zipWithM (leqSize Mixed) mu1' mu2' 
      return ()
    Pos   -> do
      lexSizes False mu2  mu1  -- not strictly smaller 
      lexSizes False mu1' mu2' 
      return ()
    Neg   -> do
      lexSizes False mu1  mu2 
      lexSizes False mu2' mu1' 
      return ()
  
{-
eqGuard :: Bound Val -> Bound Val -> TypeCheck ()
eqGuard (Bound (Measure mu1) (Measure mu1')) (Bound (Measure mu2) (Measure mu2')) = do
  zipWithM (leqSize Mixed) mu1 mu2
  zipWithM (leqSize Mixed) mu1' mu2'
  return ()
-}           